# -*- coding: utf-8 -*-
import array
import os
from struct import pack, unpack
from io import BytesIO

import pulsar
from pulsar.utils.httpurl import ispy3k, range, to_bytes, is_string


__all__ = ['WebSocketError',
           'WebSocketProtocolError',
           'Frame',
           'FrameParser',
           'i2b',
           'int2bytes']

DEFAULT_VERSION = 13
SUPPORTED_VERSIONS = (DEFAULT_VERSION,)

if ispy3k:
    i2b = lambda n : bytes((n,))
    
    def is_text_data(data):
        return isinstance(data, str)
            
else: # pragma : nocover
    i2b = lambda n : chr(n)
    
    def is_text_data(data):
        return True

def int2bytes(*ints):
    '''convert a series of integers into bytes'''
    return b''.join((i2b(i) for i in ints))

def get_version(version):
    version = int(version or DEFAULT_VERSION)
    if version not in SUPPORTED_VERSIONS:
        raise WebSocketProtocolError('Version %s not supported.' % version)
    return version


class WebSocketError(pulsar.HttpException):
    pass


class WebSocketProtocolError(WebSocketError):
    pass
    
    
class Frame(object):
    __slots__ = ('fin', 'opcode', 'masking_key', 'rsv1', 'rsv2', 'rsv3',
                 'msg', 'version', 'payload_length', 'body')
    def __init__(self, message=None, opcode=None, version=None,
                 masking_key=None, final=False, rsv1=0, rsv2=0, rsv3=0):
        """Implements the framing protocol as defined by hybi_
specification supporting protocol version 13::
    
    >>> f = Frame(opcode, 'hello world', os.urandom(4), fin=1)
    >>> data = f.build()
    >>> f = Frame()
    >>> f.parser.send(bytes[1])
    >>> f.parser.send(bytes[2])
    >>> f.parser.send(bytes[2:])
"""
        self.version = get_version(version)
        if opcode is None and message is not None:
            opcode = 0x1 if is_text_data(message) else 0x2
        message = to_bytes(message or b'')
        self.payload_length = len(message)
        if opcode is None:
            raise WebSocketProtocolError('opcode not available')
        self.version = version
        self.opcode = opcode
        if masking_key:
            masking_key = to_bytes(masking_key)
            if len(masking_key) != 4:
                raise WebSocketProtocolError('Masking key must be 4 bytes long')
        self.masking_key = masking_key
        self.fin = 0x1 if final else 0
        self.rsv1 = rsv1
        self.rsv2 = rsv2
        self.rsv3 = rsv3
        self.body = message
        self.msg = self.build_frame(message)
        
    @classmethod
    def ping(cls, body=None, version=None):
        return cls(body, 0x9, final=True, version=version)
    
    @classmethod
    def pong(cls, body=None, version=None):
        return cls(body, 0xA, final=True, version=version)
    
    @classmethod
    def close(cls, body=None, version=None):
        return cls(body, 0x8, final=True, version=version)
    
    @classmethod
    def continuation(cls, body=None, **kwargs):
        return cls(body, 0, **kwargs)
    
    @property
    def final(self):
        return bool(self.fin)
    
    @property
    def masked(self):
        return bool(self.masking_key)
    
    @property
    def is_data(self):
        return self.opcode in (0x1, 0x2)
    
    @property
    def is_close(self):
        return self.opcode == 0x8
    
    def on_received(self):
        if self.opcode == 0x9:
            return self.pong(self.body, self.version)
    
    def build_frame(self, message):
        """Builds a frame from the instance's attributes."""
        header = BytesIO()
        if 0x3 <= self.opcode <= 0x7 or 0xB <= self.opcode:
            raise WebSocketProtocolError('Opcode cannot be a reserved opcode')
        ## +-+-+-+-+-------+
        ## |F|R|R|R| opcode|
        ## |I|S|S|S|  (4)  |
        ## |N|V|V|V|       |
        ## | |1|2|3|       |
        ## +-+-+-+-+-------+
        header.write(i2b(((self.fin << 7)
                       | (self.rsv1 << 6)
                       | (self.rsv2 << 5)
                       | (self.rsv3 << 4)
                       | self.opcode)))
        ##                 +-+-------------+-------------------------------+
        ##                 |M| Payload len |    Extended payload length    |
        ##                 |A|     (7)     |             (16/63)           |
        ##                 |S|             |   (if payload len==126/127)   |
        ##                 |K|             |                               |
        ## +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
        ## |     Extended payload length continued, if payload len == 127  |
        ## + - - - - - - - - - - - - - - - +-------------------------------+
        if self.masking_key:
            mask_bit = 1 << 7
        else:
            mask_bit = 0
        length = self.payload_length 
        if length < 126:
            header.write(i2b(mask_bit | length))
        elif length < (1 << 16):
            header.write(i2b(mask_bit | 126))
            header.write(pack('!H', length))
        elif length < (1 << 63):
            header.write(i2b(mask_bit | 127))
            header.write(pack('!Q', length))
        else:
            raise WebSocketProtocolError('Frame too large')
        ## + - - - - - - - - - - - - - - - +-------------------------------+
        ## |                               |Masking-key, if MASK set to 1  |
        ## +-------------------------------+-------------------------------+
        ## | Masking-key (continued)       |          Payload Data         |
        ## +-------------------------------- - - - - - - - - - - - - - - - +
        ## :                     Payload Data continued ...                :
        ## + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        ## |                     Payload Data continued ...                |
        ## +---------------------------------------------------------------+
        if not self.masking_key:
            header.write(message)
        else:
            header.write(self.masking_key)
            header.write(self.mask(message))
        return header.getvalue()
                
    def mask(self, data):
        """Performs the masking or unmasking operation on data
using the simple masking algorithm::

    j                   = i MOD 4
    transformed-octet-i = original-octet-i XOR masking-key-octet-j
"""
        masked = bytearray(data)
        key = self.masking_key
        if not ispy3k:  #pragma    nocover
            key = map(ord, key)
        for i in range(len(data)):
            masked[i] = masked[i] ^ key[i%4]
        return masked
    unmask = mask


class ParsedFrame(Frame):
    def build_frame(self, msg):
        return bytearray(msg)
    
    
class FrameParser(object):
    '''Parser for the version 8 protocol.
    
.. attribute:: kind

    0 for parsing client's frames and sending server frames (to be used in the server),
    1 for parsing server frames and sending client frames (to be used by the client)
'''    
    def __init__(self, version=None, kind=0):
        self.version = get_version(version)
        self._frame = None
        self._buf = None
        self._kind = kind
        
    @property
    def kind(self):
        return self._kind
    
    @property
    def masked(self):
        return self.kind == 1
    
    @property
    def expect_masked(self):
        return True if self.kind == 0 else False
    
    def decode(self, data):
        return self.execute(data), bytearray()
    
    def encode(self, data):
        # Encode data into a frame
        if not isinstance(data, Frame):
            if self.kind == 1:
                data = Frame(data, masking_key=os.urandom(4))
            else:
                data = Frame(data)
        return data.msg
    
    def execute(self, data, length=None):
        # end of body can be passed manually by putting a length of 0
        if not data:
            return None
        data = bytes(data)
        frame = self._frame
        if self._buf:
            data = self._buf + data
        masked_frame = self.expect_masked
        # No opcode yet
        if frame is None:
            if len(data) < 2:
                return self.save_buf(frame, data)
            first_byte, second_byte = unpack("BB", data[:2])
            fin = (first_byte >> 7) & 1
            rsv1 = (first_byte >> 6) & 1
            rsv2 = (first_byte >> 5) & 1
            rsv3 = (first_byte >> 4) & 1
            opcode = first_byte & 0xf
            if fin not in (0, 1):
                raise WebSocketProtocolError('FIN must be 0 or 1')
            is_masked = bool(second_byte & 0x80)
            if masked_frame != is_masked:
                if masked_frame:
                    raise WebSocketProtocolError('WEBSOCKET unmasked client frame.')
                else:
                    raise WebSocketProtocolError('WEBSOCKET masked server frame.')
            payload_length = second_byte & 0x7f
            # All control frames MUST have a payload length of 125 bytes or less
            if opcode > 0x7 and payload_length > 125:
                raise WebSocketProtocolError('Frame too large')
            msg, data = data[:2], data[2:]
            frame = ParsedFrame(msg, opcode, self.version, final=fin,
                                rsv1=rsv1, rsv2=rsv2, rsv3=rsv3)
            frame.payload_length = payload_length
        #   
        if frame.masking_key is None:
            # All control frames MUST have a payload length of 125 bytes or less
            d = None
            mask_length = 4 if masked_frame else 0
            if frame.payload_length == 126:
                if len(data) < 2 + mask_length: # 2 + 4 for mask
                     return self.save_buf(frame, data)
                d, data = data[:2] , data[2:]
                frame.payload_length = unpack("!H", d)[0]
            elif frame.payload_length == 127:
                if len(data) < 8 + mask_length:  # 8 + 4 for mask
                     return self.save_buf(frame, data)
                d, data = data[:8] , data[8:]
                frame.payload_length = unpack("!Q", d)[0]
            elif len(data) < mask_length:
                return self.save_buf(frame, data)
            if d:
                frame.msg.extend(d)
            if mask_length:
                frame.masking_key, data = data[:mask_length], data[mask_length:]
            else:
                frame.masking_key = b''
            frame.msg.extend(frame.masking_key)
                
        if len(data) < frame.payload_length:
            return self.save_buf(frame, data)
        # We have a frame
        else:
            data = data[:frame.payload_length]
            frame.msg.extend(data)
            self.save_buf(None, data[frame.payload_length:])
            if frame.masking_key:
                data = bytes(frame.unmask(data))
            if frame.opcode == 0x1:
                data = data.decode("utf-8", "replace")
            frame.body = data
            return frame
            
    def save_buf(self, frame, data):
        self._frame = frame
        self._buf = data
        


import sys
import time
import os
import logging
import socket
from wsgiref.handlers import format_date_time
from io import BytesIO

import pulsar
from pulsar import lib, make_async, is_async, AsyncSocketServer, Deferred,\
                   AsyncConnection, AsyncResponse, HttpException
from pulsar.utils.httpurl import Headers, is_string, unquote,\
                                    has_empty_content, to_bytes,\
                                    host_and_port_default, mapping_iterator,\
                                    Headers, REDIRECT_CODES
from pulsar.utils import events

from .wsgi import WsgiResponse, handle_wsgi_error


__all__ = ['HttpServer']


def wsgi_environ(connection, parser):
    """return a :ref:`WSGI <apps-wsgi>` compatible environ dictionary
based on the current request. If the reqi=uest headers are not ready it returns
nothing."""
    version = parser.get_version()
    input = BytesIO()
    for b in parser.get_body():
        input.write(b)
    input.seek(0)
    protocol = parser.get_protocol()
    environ = {
        "wsgi.input": input,
        "wsgi.errors": sys.stderr,
        "wsgi.version": version,
        "wsgi.run_once": True,
        "wsgi.url_scheme": protocol,
        "SERVER_SOFTWARE": pulsar.SERVER_SOFTWARE,
        "REQUEST_METHOD": parser.get_method(),
        "QUERY_STRING": parser.get_query_string(),
        "RAW_URI": parser.get_url(),
        "SERVER_PROTOCOL": protocol,
        'CONTENT_TYPE': '',
        "CONTENT_LENGTH": '',
        'SERVER_NAME': connection.server.server_name,
        'SERVER_PORT': connection.server.server_port,
        "wsgi.multithread": False,
        "wsgi.multiprocess":False
    }
    # REMOTE_HOST and REMOTE_ADDR may not qualify the remote addr:
    # http://www.ietf.org/rfc/rfc3875
    url_scheme = "http"
    forward = connection.address
    server = None
    url_scheme = "http"
    script_name = os.environ.get("SCRIPT_NAME", "")
    headers = mapping_iterator(parser.get_headers())
    for header, value in headers:
        header = header.lower()
        if header == 'x-forwarded-for':
            forward = value
        elif header == "x-forwarded-protocol" and value == "ssl":
            url_scheme = "https"
        elif header == "x-forwarded-ssl" and value == "on":
            url_scheme = "https"
        elif header == "host":
            server = value
        elif header == "script_name":
            script_name = value
        elif header == "content-type":
            environ['CONTENT_TYPE'] = value
            continue
        elif header == "content-length":
            environ['CONTENT_LENGTH'] = value
            continue
        key = 'HTTP_' + header.upper().replace('-', '_')
        environ[key] = value
    environ['wsgi.url_scheme'] = url_scheme
    if is_string(forward):
        # we only took the last one
        # http://en.wikipedia.org/wiki/X-Forwarded-For
        if forward.find(",") >= 0:
            forward = forward.rsplit(",", 1)[1].strip()
        remote = forward.split(":")
        if len(remote) < 2:
            remote.append('80')
    else:
        remote = forward
    environ['REMOTE_ADDR'] = remote[0]
    environ['REMOTE_PORT'] = str(remote[1])
    if server is not None:
        server =  host_and_port_default(url_scheme, server)
        environ['SERVER_NAME'] = server[0]
        environ['SERVER_PORT'] = server[1]
    path_info = parser.get_path()
    if path_info is not None:
        if script_name:
            path_info = path_info.split(script_name, 1)[1]
        environ['PATH_INFO'] = unquote(path_info)
    environ['SCRIPT_NAME'] = script_name
    return environ

def chunk_encoding(chunk):
    '''Write a chunk::

    chunk-size(hex) CRLF
    chunk-data CRLF
    
If the size is 0, this is the last chunk, and an extra CRLF is appended.
'''
    head = ("%X\r\n" % len(chunk)).encode('utf-8')
    return head + chunk + b'\r\n'


class HttpResponse(AsyncResponse):
    '''Handle an HTTP response for a :class:`HttpConnection`.'''
    _status = None
    _headers_sent = None
    headers = None
    MAX_CHUNK = 65536

    def default_headers(self):
        return Headers([('Server', pulsar.SERVER_SOFTWARE),
                        ('Date', format_date_time(time.time()))])

    @property
    def environ(self):
        return self.parsed_data

    @property
    def status(self):
        return self._status

    @property
    def upgrade(self):
        if self.headers:
            return self.headers.get('Upgrade')

    @property
    def chunked(self):
        return self.headers.get('Transfer-Encoding') == 'chunked'

    @property
    def content_length(self):
        c = self.headers.get('Content-Length')
        if c:
            return int(c)

    @property
    def version(self):
        return self.environ.get('wsgi.version')

    @property
    def keep_alive(self):
        """ return True if the connection should be kept alive"""
        conn = self.environ.get('HTTP_CONNECTION','').lower()
        if conn == "close":
            return False
        elif conn == "keep-alive":
            return True
        return self.version == (1, 1)

    def start_response(self, status, response_headers, exc_info=None):
        '''WSGI compliant ``start_response`` callable, see pep3333_.
The application may call start_response more than once, if and only
if the exc_info argument is provided.
More precisely, it is a fatal error to call start_response without the exc_info
argument if start_response has already been called within the current
invocation of the application.

:parameter status: an HTTP "status" string like "200 OK" or "404 Not Found".
:parameter response_headers: a list of ``(header_name, header_value)`` tuples.
    It must be a Python list. Each header_name must be a valid HTTP header
    field-name (as defined by RFC 2616_, Section 4.2), without a trailing
    colon or other punctuation.
:parameter exc_info: optional python ``sys.exc_info()`` tuple. This argument
    should be supplied by the application only if start_response is being
    called by an error handler.

:rtype: The :meth:`HttpResponse.write` callable.

.. _pep3333: http://www.python.org/dev/peps/pep-3333/
.. _2616: http://www.faqs.org/rfcs/rfc2616.html
'''
        if exc_info:
            try:
                if self._headers_sent:
                    # if exc_info is provided, and the HTTP headers have
                    # already been sent, start_response must raise an error,
                    # and should re-raise using the exc_info tuple
                    raise (exc_info[0], exc_info[1], exc_info[2])
            finally:
                # Avoid circular reference
                exc_info = None
        elif self._status:
            # Headers already sent. Raise error
            raise pulsar.HttpException("Response headers already sent!")
        self._status = status
        if type(response_headers) is not list:
            raise TypeError("Headers must be a list of name/value tuples")
        self.headers = self.default_headers()
        self.headers.update(response_headers)
        return self.write

    def write(self, data):
        '''The write function required by WSGI specification.'''
        head = self.send_headers(force=data)
        if head:
            self.connection.write(head)
        if data:
            self.connection.write(data)

    def _generate(self, response):
        MAX_CHUNK = self.MAX_CHUNK
        for b in response:
            head = self.send_headers(force=b)
            if head is not None:
                yield head
            if b:
                if self.chunked:
                    while len(b) >= MAX_CHUNK:
                        chunk, b = b[:MAX_CHUNK], b[MAX_CHUNK:]
                        yield chunk_encoding(chunk)
                    if b:
                        yield chunk_encoding(b)
                else:
                    yield b
            else:
                yield b''
            
    def __iter__(self):
        conn = self.connection
        keep_alive = self.keep_alive
        self.environ['pulsar.connection'] = self.connection
        try:
            resp = conn.server.app_handler(self.environ, self.start_response)
            for b in self._generate(resp):
                yield b
        except Exception as e:
            exc_info = sys.exc_info()
            if self._headers_sent:
                keep_alive = False
                conn.log.critical('Headers already sent', exc_info=exc_info)
                yield b'CRITICAL SERVER ERROR. Please Contact the administrator'
            else:
                # Create the error response
                resp = handle_wsgi_error(self.environ, exc_info)
                keep_alive = keep_alive and resp.status_code in REDIRECT_CODES
                resp.headers['connection'] =\
                    "keep-alive" if keep_alive else "close"
                resp(self.environ, self.start_response, exc_info)
                for b in self._generate(resp):
                    yield b
        # make sure we send the headers
        head = self.send_headers(force=True)
        if head is not None:
            yield head
        if self.chunked:
            # Last chunk
            yield chunk_encoding(b'')
        # close connection if required
        if not keep_alive:
            self.connection.close()

    def is_chunked(self):
        '''Only use chunked responses when the client is
speaking HTTP/1.1 or newer and there was no Content-Length header set.'''
        if self.environ['wsgi.version'] <= (1,0):
            return False
        elif has_empty_content(int(self.status[:3])):
            # Do not use chunked responses when the response
            # is guaranteed to not have a response body.
            return False
        elif self.headers.get('Transfer-Encoding') == 'chunked':
            return True
        else:
            return self.content_length is None

    def get_headers(self, force=False):
        '''Get the headers to send only if *force* is ``True`` or this
is an HTTP upgrade (websockets)'''
        if self.upgrade or force:
            if not self._status:
                # we are sending headers but the start_response was not called
                raise HttpException('Headers not set.')
            headers = self.headers
            # Set chunked header if needed
            if self.is_chunked():
                headers['Transfer-Encoding'] = 'chunked'
                headers.pop('content-length', None)
            else:
                headers.pop('Transfer-Encoding', None)
            if 'connection' not in headers:
                connection = "keep-alive" if self.keep_alive else "close"
                headers['Connection'] = connection
            return headers

    def send_headers(self, force=False):
        if not self._headers_sent:
            tosend = self.get_headers(force)
            if tosend:
                events.fire('http-headers', self, headers=tosend)
                self._headers_sent = tosend.flat(self.version, self.status)
                return self._headers_sent

    def close(self, msg=None):
        self.connection._current_response = None
        if not self.keep_alive:
            return self.connection.close(msg)


class HttpProtocol:
    connection = None
    def __init__(self):
        self.parser = lib.Http_Parser(kind=0)
        
    def decode(self, data):
        p = self.parser
        if p.is_message_complete():
            self.parser = p = lib.Http_Parser(kind=0)
        if data:
            headers = self.headers()
            if p.execute(bytes(data), len(data)) == len(data):
                if headers is None:
                    self.expect_continue()
                if p.is_message_complete():
                    return wsgi_environ(self.connection, p), bytearray()
        return None, bytearray()
    
    def headers(self):
        if self.parser.is_headers_complete():
            return self.parser.get_headers()
    
    def expect_continue(self):
        '''Handle the expect=100-continue header if available, according to
the following algorithm:

* Send the 100 Continue response before waiting for the body.
* Omit the 100 (Continue) response if it has already received some or all of
  the request body for the corresponding request.
    '''
        headers = self.headers()
        if headers is not None and headers.get('Expect') == '100-continue':
            if not self.parser.is_message_complete():
                self.connection.write(b'HTTP/1.1 100 Continue\r\n\r\n')


class HttpServer(AsyncSocketServer):
    response_class = HttpResponse
    protocol_factory = HttpProtocol

    def __init__(self, *args, **kwargs):
        super(HttpServer, self).__init__(*args, **kwargs)
        host, port = self.sock.getsockname()[:2]
        self.server_name = socket.getfqdn(host)
        self.server_port = port
        self.app_handler = self.actor.app_handler
    




"""A specialized :class:`pulsar.apps.socket.SocketServer` for
serving web applications which conforms with the python web server
gateway interface (WSGI_).

The application can be used in conjunction with several web frameworks
as well as the :ref:`pulsar RPC middleware <apps-rpc>` and
the :ref:`websocket middleware <apps-ws>`.

An example of a web server written with ``pulsar.apps.wsgi`` which responds
with "Hello World!" for every request::

    from pulsar.apps import wsgi

    def hello(environ, start_response):
        data = b"Hello World!"
        response_headers = (
            ('Content-type','text/plain'),
            ('Content-Length', str(len(data)))
        )
        start_response("200 OK", response_headers)
        return [data]

    if __name__ == '__main__':
        wsgi.WSGIServer(callable=hello).start()


For more information regarding WSGI check the pep3333_ specification.

.. _pep3333: http://www.python.org/dev/peps/pep-3333/
.. _WSGI: http://www.wsgi.org
"""
import sys
from inspect import isclass

import pulsar
from pulsar.utils.importer import module_attribute
from pulsar.apps import socket

from .wsgi import *
from .server import *
from .middleware import *
from .request import *


class WsgiSetting(pulsar.Setting):
    virtual = True
    app = 'wsgi'
    section = "WSGI Servers"


class HttpParser(WsgiSetting):
    name = "http_parser"
    flags = ["--http-parser"]
    desc = """\
        The HTTP Parser to use. By default it uses the fastest possible.

        Specify `python` if you wich to use the pure python implementation
        """


class ResponseMiddleware(WsgiSetting):
    name = "response_middleware"
    flags = ["--response-middleware"]
    nargs = '*'
    desc = """\
    Response middleware to add to the wsgi handler
    """


class WSGIServer(socket.SocketServer):
    cfg_apps = ('socket',)
    _app_name = 'wsgi'
    socket_server_factory = HttpServer

    def handler(self):
        '''Build the wsgi handler from *hnd*. This function is called
at start-up only.

:parameter hnd: This is the WSGI handle which can be A :class:`WsgiHandler`,
    a WSGI callable or a list WSGI callables.
:parameter resp_middleware: Optional list of response middleware functions.'''
        hnd = self.callable
        if not isinstance(hnd, WsgiHandler):
            if not isinstance(hnd, (list, tuple)):
                hnd = [hnd]
            hnd = WsgiHandler(hnd)
        response_middleware = self.cfg.response_middleware or []
        for m in response_middleware:
            if '.' not in m:
                mm = getattr(middleware,m,None)
                if not mm:
                    raise ValueError('Response middleware "{0}" not available'\
                                     .format(m))
            else:
                mm = module_attribute(m)
            if isclass(mm):
                mm = mm()
            hnd.response_middleware.append(mm)
        return hnd



'''The :mod:`pulsar.apps.ws` contains a WSGI_ middleware for
handling the WebSocket_ protocol.
Web sockets allow for bidirectional communication between the browser
and server. Pulsar implementation uses the WSGI middleware
:class:`WebSocket` for the handshake and a class derived from
:class:`WS` handler for the communication part.

.. _WSGI: http://www.python.org/dev/peps/pep-3333/
.. _WebSocket: http://tools.ietf.org/html/rfc6455

API
==============

WebSocket
~~~~~~~~~~~~~~~~

.. autoclass:: WebSocket
   :members:
   :member-order: bysource


WebSocket Handler
~~~~~~~~~~~~~~~~~~~~

.. autoclass:: WS
   :members:
   :member-order: bysource

Frame
~~~~~~~~~~~~~~~~~~~

.. autoclass:: Frame
   :members:
   :member-order: bysource
   
   
.. autoclass:: FrameParser
   :members:
   :member-order: bysource


'''
import logging
import socket
import re
import struct
import time
import base64
import hashlib
from functools import partial

import pulsar
from pulsar import maybe_async, is_async, safe_async, is_failure, ClientSocket
from pulsar.utils.httpurl import ispy3k, to_bytes, native_str,\
                                 itervalues, parse_qs, WEBSOCKET_VERSION
from pulsar.apps.wsgi import WsgiResponse, wsgi_iterator

from .frame import *

WEBSOCKET_GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'


class GeneralWebSocket(object):
    namespace = ''
    extensions = []
    
    def __init__(self, handle, extensions=None):
        self.handle = handle
        if extensions is None:
            extensions = self.extensions
        self.extensions = extensions

    def get_client(self):
        return self.handle(self)
    
    def __call__(self, environ, start_response):
        if self.handle.match(environ):
            return self.handle_handshake(environ, start_response)
    
    def handle_handshake(self, environ, start_response):
        raise NotImplementedError()
    
    def upgrade_connection(self, environ, version):
        '''Upgrade the connection so it handles the websocket protocol.'''
        connection = environ['pulsar.connection']
        connection.environ = environ
        #connection.read_timeout = 0
        connection.protocol = FrameParser(version)
        connection.response_class = self.on_message
        
    def on_message(self, connection, frame):
        environ = connection.environ
        if not environ.get('websocket-opened'):
            environ['websocket-opened'] = True
            self.handle.on_open(environ)
        rframe = frame.on_received()
        if rframe:
            yield rframe.msg
        elif frame.is_close:
            # Close
            connection.close()
        elif frame.is_data:
            yield self.as_frame(connection,
                                self.handle.on_message(environ, frame.body))

    def as_frame(self, connection, body):
        '''Build a websocket server frame from body.'''
        body = maybe_async(body)
        if is_async(body):
            return body.addBoth(lambda b: self.as_frame(connection, b))
        if is_failure(body):
            # We have a failure. shut down connection
            body.log()
            body = Frame.close('Server error')
        elif not isinstance(body, Frame):
            # If the body is not a frame, build a final frame from it.
            body = Frame(body or '', version=connection.protocol.version,
                         final=True)
        return body.msg
        
    
class WebSocket(GeneralWebSocket):
    """A :ref:`WSGI <apps-wsgi>` middleware for handling w websocket handshake
and starting a custom :class:`WS` connection.
It implements the protocol version 13 as specified at
http://www.whatwg.org/specs/web-socket-protocol/.

Web Sockets are not standard HTTP connections. The "handshake" is HTTP,
but after that, the protocol is message-based. To create
a valid :class:`WebSocket` middleware initialise as follow::

    from pulsar.apps import wsgi, ws
    
    class MyWS(ws.WS):
        ...
    
    wm = ws.WebSocket(handle=MyWS())
    app = wsgi.WsgiHandler(middleware=(..., wm))
    
    wsgi.createServer(callable=app).start()


See http://tools.ietf.org/html/rfc6455 for the websocket server protocol and
http://www.w3.org/TR/websockets/ for details on the JavaScript interface.
    """        
    def handle_handshake(self, environ, start_response):
        connections = environ.get("HTTP_CONNECTION", '').lower()\
                                    .replace(' ','').split(',')
        if environ.get("HTTP_UPGRADE", '').lower() != "websocket" or \
           'upgrade' not in connections:
            return
        if environ['REQUEST_METHOD'].upper() != 'GET':
            raise WebSocketError('Method is not GET', status=400)
        key = environ.get('HTTP_SEC_WEBSOCKET_KEY')
        if key:
            try:
                ws_key = base64.b64decode(key.encode('latin-1'))
            except:
                ws_key = ''
            if len(ws_key) != 16:
                raise WebSocketError("WebSocket key's length is invalid",
                                     status=400)
        else:
            raise WebSocketError('Not a valid HyBi WebSocket request. '
                                 'Missing Sec-Websocket-Key header.',
                                 status=400)
        version = environ.get('HTTP_SEC_WEBSOCKET_VERSION')
        if version:
            try:
                version = int(version)
            except:
                pass
        if version not in WEBSOCKET_VERSION:
            raise WebSocketError('Unsupported WebSocket version {0}'\
                                 .format(version),
                                 status=400)
        # Collect supported subprotocols
        subprotocols = environ.get('HTTP_SEC_WEBSOCKET_PROTOCOL')
        ws_protocols = []
        if subprotocols:
            for s in subprotocols.split(','):
                s = s.strip()
                if s in protocols:
                    ws_protocols.append(s)
        # Collect supported extensions
        ws_extensions = []
        extensions = environ.get('HTTP_SEC_WEBSOCKET_EXTENSIONS')
        if extensions:
            exts = self.extensions
            for ext in extensions.split(','):
                ext = ext.strip()
                if ext in exts:
                    ws_extensions.append(ext)
        # Build and start the HTTP response
        headers = [
            ('Upgrade', 'websocket'),
            ('Connection', 'Upgrade'),
            ('Sec-WebSocket-Accept', self.challenge_response(key))
        ]
        if ws_protocols:
            headers.append(('Sec-WebSocket-Protocol',
                            ', '.join(ws_protocols)))
        if ws_extensions:
            headers.append(('Sec-WebSocket-Extensions',
                            ','.join(ws_extensions)))
        self.handle.on_handshake(environ, headers)
        self.upgrade_connection(environ, version)
        response = WsgiResponse(101, content=(), response_headers=headers)
        return response(environ, start_response)
        
    def challenge_response(self, key):
        sha1 = hashlib.sha1(to_bytes(key+WEBSOCKET_GUID))
        return native_str(base64.b64encode(sha1.digest()))
    
    def get_parser(self):
        return FrameParser()


class WS(object):
    '''A web socket handler. It maintains
an open socket with a remote web-socket and exchange messages in
an asynchronous fashion. The communication is started by the
:class:`WebSocket` middleware after a successful handshake.
 
Override :meth:`on_message` to handle incoming messages.
You can also override :meth:`on_open` and :meth:`on_close` to handle opened
and closed connections.

Here is an example Web Socket handler that echos back all received messages
back to the client::

    from pulsar.apps import ws
    
    class EchoWS(ws.WS):
    
        def on_open(self, environ):
            print("WebSocket opened")
    
        def on_message(self, environ, message):
            return message
    
        def on_close(self, environ):
            print("WebSocket closed")
            
'''
    def match(self, environ):
        pass
    
    def on_handshake(self, environ, headers):
        """Invoked just before sending the upgraded **headers** to the
client. This is a chance to add or remove header's entries."""
        pass
    
    def on_open(self, environ):
        """Invoked when a new WebSocket is opened."""
        pass

    def on_message(self, environ, message):
        """Handle incoming messages on the WebSocket.
        This method must be overloaded.
        """
        raise NotImplementedError()

    def on_close(self, environ):
        """Invoked when the WebSocket is closed."""
        pass
    
    def close(self, environ, msg=None):
        '''Invoked when the web-socket needs closing.'''
        connection = environ['pulsar.connection']
        return Frame.close(msg, version=connection.protocol.version)
    
    
class WebSocketClient(ClientSocket):
    
    def isclosed(self):
        # For compatibility with HttpResponse
        return self.closed
    
    def protocol_factory(self):
        return FrameParser(kind=1)
    

class HttpClient(pulsar.HttpClient):
    
    def upgrade(self, response):
        client = WebSocketClient(response.sock, response.url)
        client.handshake = response
        return client
        


'''Tests the "helloworld" example.'''
from pulsar import send, SERVER_SOFTWARE, HttpClient, get_application
from pulsar import MultiDeferred
from pulsar.utils.httpurl import range
from pulsar.apps.test import unittest, run_on_arbiter, dont_run_with_thread

from .manage import server
        

class TestHelloWorldThread(unittest.TestCase):
    app = None
    concurrency = 'thread'
    
    @classmethod
    def name(cls):
        return 'helloworld_' + cls.concurrency
    
    @classmethod
    def setUpClass(cls):
        name = name=cls.name()
        kwargs = {'%s__bind' % name: '127.0.0.1:0'}
        s = server(name=cls.name(), concurrency=cls.concurrency, **kwargs)
        outcome = send('arbiter', 'run', s)
        yield outcome
        cls.app = outcome.result
        cls.uri = 'http://{0}:{1}'.format(*cls.app.address)
        
    @classmethod
    def tearDownClass(cls):
        if cls.app is not None:
            outcome = send('arbiter', 'kill_actor', cls.app.name)
            yield outcome
    
    @run_on_arbiter
    def testMeta(self):
        app = get_application(self.name())
        self.assertEqual(app.name, self.name())
        self.assertTrue(app.monitor.running())
        self.assertEqual(app, app.app)
        self.assertEqual(str(app), app.name)
        self.assertEqual(app.cfg.bind, '127.0.0.1:0')
        
    def testResponse(self):
        c = HttpClient()
        outcome = c.get(self.uri)
        yield outcome
        resp = outcome.result
        self.assertEqual(resp.status_code, 200)
        content = resp.content
        self.assertEqual(content, b'Hello World!\n')
        headers = resp.headers
        self.assertTrue(headers)
        self.assertEqual(headers['content-type'], 'text/plain')
        self.assertEqual(headers['server'], SERVER_SOFTWARE)
        
    def test_getbench(self):
        c = HttpClient()
        yield MultiDeferred((c.get(self.uri) for _ in range(1))).lock()
    test_getbench.__benchmark__ = True


@dont_run_with_thread
class TestHelloWorldProcess(TestHelloWorldThread):
    concurrency = 'process'
    


'''Response middleware
'''
import re
import os
import stat
import mimetypes
from email.utils import parsedate_tz, mktime_tz
from gzip import GzipFile

import pulsar
from pulsar.utils.httpurl import BytesIO, parse_authorization_header,\
                                     parse_cookie, http_date
                                     
from .wsgi import WsgiResponse

re_accepts_gzip = re.compile(r'\bgzip\b')


__all__ = ['clean_path_middleware',
           'AccessControl',
           'GZipMiddleware',
           'cookies_middleware',
           'authorization_middleware',
           'MediaMiddleware']


def clean_path_middleware(environ, start_response):
    '''Clean url and redirect if needed'''
    path = environ['PATH_INFO']
    if '//' in path:
        url = re.sub("/+" , '/', path)
        if not url.startswith('/'):
            url = '/%s' % url
        qs = environ['QUERY_STRING']
        if qs:
            url = '%s?%s' % (url, qs)
        raise pulsar.HttpRedirect(url)

def cookies_middleware(environ, start_response):
    '''Parse the ``HTTP_COOKIE`` key in the *environ*. The ``HTTP_COOKIE``
string is replaced with a dictionary.'''
    c = environ.get('HTTP_COOKIE', '')
    if not isinstance(c, dict):
        if not c:
            c = {}
        else:
            if not isinstance(c, str):
                c = c.encode('utf-8')
            c = parse_cookie(c)
        environ['HTTP_COOKIE'] = c

def authorization_middleware(environ, start_response):
    """An `Authorization` middleware."""
    code = 'HTTP_AUTHORIZATION'
    if code in environ:
        environ[code] = parse_authorization_header(environ[code])

#####################################################    RESPONSE MIDDLEWARE
class ResponseMiddleware(object):

    def version(self, environ):
        return environ.get('wsgi.version')

    def available(self, environ, response):
        return True

    def __call__(self, environ, response):
        if not self.available(environ, response):
            return
        return self.execute(environ, response)

    def execute(self, environ, response):
        pass


class AccessControl(ResponseMiddleware):
    '''A response middleware which add the ``Access-Control-Allow-Origin``
response header.'''
    def __init__(self, origin='*', methods=None):
        self.origin = origin
        self.methods = methods

    def available(self, environ, response):
        return response.status_code == 200

    def execute(self, environ, response):
        response.headers['Access-Control-Allow-Origin'] = self.origin
        if self.methods:
            response.headers['Access-Control-Allow-Methods'] = self.methods


class GZipMiddleware(ResponseMiddleware):
    """Response middleware for compressing content if the browser allows
gzip compression. It sets the Vary header accordingly, so that caches will
base their storage on the Accept-Encoding header.
    """
    def __init__(self, min_length=200):
        self.min_length = min_length

    def available(self, environ, response):
        # It's not worth compressing non-OK or really short responses
        if response.status_code == 200 and not response.is_streamed:
            if response.length() < self.min_length:
                return False
            headers = response.headers
            # Avoid gzipping if we've already got a content-encoding.
            if 'Content-Encoding' in headers:
                return False
            # MSIE have issues with gzipped response of various content types.
            if "msie" in environ.get('HTTP_USER_AGENT', '').lower():
                ctype = headers.get('Content-Type', '').lower()
                if not ctype.startswith("text/") or "javascript" in ctype:
                    return False
            ae = environ.get('HTTP_ACCEPT_ENCODING', '')
            if not re_accepts_gzip.search(ae):
                return False
            return True

    def execute(self, environ, response):
        headers = response.headers
        headers.add_header('Vary', 'Accept-Encoding')
        content = b''.join(response.content)
        response.content = (self.compress_string(content),)
        response.headers['Content-Encoding'] = 'gzip'

    # From http://www.xhaus.com/alan/python/httpcomp.html#gzip
    def compress_string(self, s):
        zbuf = BytesIO()
        zfile = GzipFile(mode='wb', compresslevel=6, fileobj=zbuf)
        zfile.write(s)
        zfile.close()
        return zbuf.getvalue()


class MediaMiddleware:
    DEFAULT_CONTENT_TYPE = 'text/css'
    def __init__(self, dir_path, url_path='/media/'):
        self.dir_path = dir_path
        self.url_path = url_path
        
    def __call__(self, environ, start_response):
        path = environ.get('PATH_INFO')
        if path.startswith(self.url_path):
            path = self.dir_path + path[len(self.url_path):]
            if os.path.isfile(path):
                response = self.serve_file(environ, path)
                return response(environ, start_response)
        
    def serve_file(self, environ, fullpath):
        # Respect the If-Modified-Since header.
        statobj = os.stat(fullpath)
        content_type, encoding = mimetypes.guess_type(fullpath)
        content_type = content_type or self.DEFAULT_CONTENT_TYPE
        if not self.was_modified_since(environ.get('HTTP_IF_MODIFIED_SINCE'),
                                       statobj[stat.ST_MTIME],
                                       statobj[stat.ST_SIZE]):
            return wsgi.WsgiResponse(status=304,
                                     content_type=content_type,
                                     encoding=encoding)
        contents = open(fullpath, 'rb').read()
        response = wsgi.WsgiResponse(content=contents,
                                     content_type=content_type,
                                     encoding=encoding)
        response.headers["Last-Modified"] = http_date(statobj[stat.ST_MTIME])
        return response
    
    def was_modified_since(self, header=None, mtime=0, size=0):
        """
        Was something modified since the user last downloaded it?

        header
          This is the value of the If-Modified-Since header.  If this is None,
          I'll just return True.

        mtime
          This is the modification time of the item we're talking about.

        size
          This is the size of the item we're talking about.
        """
        try:
            if header is None:
                raise ValueError
            matches = re.match(r"^([^;]+)(; length=([0-9]+))?$", header,
                               re.IGNORECASE)
            header_mtime = mktime_tz(parsedate_tz(matches.group(1)))
            header_len = matches.group(3)
            if header_len and int(header_len) != size:
                raise ValueError()
            if mtime > header_mtime:
                raise ValueError()
        except (AttributeError, ValueError, OverflowError):
            return True
        return False


'''Asynchronous WSGI Remote Procedure Calls middleware. It implements a
JSON-RPC_ server and client.

RPC server
=====================
To create a server first you create your rpc handler and
(optional) subhandlers::

    from pulsar.apps import rpc, wsgi
    
    class Root(rpc.PulsarServerCommands):
        pass
        
    class Calculator(rpc.JSONRPC):
        
        def rpc_add(self, request, a, b):
            return float(a) + float(b)
        
        def rpc_subtract(self, request, a, b):
            return float(a) - float(b)
        
        def rpc_multiply(self, request, a, b):
            return float(a) * float(b)
        
        def rpc_divide(self, request, a, b):
            return float(a) / float(b)
    

Then you create the WSGI_ middleware::

    def server():
        root = Root().putSubHandler('calc',Calculator())
        return wsgi.WSGIServer(callable=rpc.RpcMiddleware(root))
    
    if __name__ == '__main__':
        server().start()
    
.. _JSON-RPC: http://www.jsonrpc.org/specification
.. _WSGI: http://www.python.org/dev/peps/pep-3333/
'''
from .exceptions import *
from .handlers import *
from .jsonrpc import *
from .decorators import *
from .mixins import *


import sys
import os
import signal
import atexit
import socket
from time import time
import random
import threading
from multiprocessing import current_process
from multiprocessing.queues import Empty
from threading import current_thread

try:    #pragma nocover
    import queue
except ImportError: #pragma nocover
    import Queue as queue
ThreadQueue = queue.Queue

from pulsar import AlreadyCalledError, AlreadyRegistered,\
                   ActorAlreadyStarted, LogginMixin, system, Config
from pulsar.utils.structures import AttributeDictionary
from pulsar.utils import events
from .eventloop import IOLoop, setid
from .proxy import ActorProxy, ActorMessage, get_command, get_proxy
from .defer import make_async, is_failure, iteritems, itervalues,\
                     pickle, async, log_failure, is_async,\
                     as_failure, EXIT_EXCEPTIONS
from .mailbox import IOQueue, mailbox
from .access import set_local_data, is_mainthread, get_actor, remove_actor


__all__ = ['is_actor', 'send', 'Actor', 'ACTOR_STATES', 'Pulsar', 'ThreadQueue']

ACTOR_STATES = AttributeDictionary(INITIAL=0X0,
                                   INACTIVE=0X1,
                                   RUN=0x2,
                                   STOPPING=0x3,
                                   CLOSE=0x4,
                                   TERMINATE=0x5)
ACTOR_STATES.DESCRIPTION = {ACTOR_STATES.INACTIVE: 'inactive',
                            ACTOR_STATES.INITIAL: 'initial',
                            ACTOR_STATES.RUN: 'running',
                            ACTOR_STATES.STOPPING: 'stopping',
                            ACTOR_STATES.CLOSE: 'closed',
                            ACTOR_STATES.TERMINATE:'terminated'}
EXIT_SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT, system.SIGQUIT)
#
# LOW LEVEL CONSTANTS - NO NEED TO CHANGE THOSE ###########################
MIN_NOTIFY = 5     # DON'T NOTIFY BELOW THIS INTERVAL
MAX_NOTIFY = 30    # NOTIFY AT LEAST AFTER THESE SECONDS
ACTOR_TERMINATE_TIMEOUT = 2 # TIMEOUT WHEN JOINING A TERMINATING ACTOR
ACTOR_TIMEOUT_TOLERANCE = 0.6
ACTOR_STOPPING_LOOPS = 5

def is_actor(obj):
    return isinstance(obj, Actor)


def send(target, action, *args, **params):
    '''Send an *message* to *target* to perform a given *action*.

:parameter target: the :class:`Actor` id or an :class:`ActorProxy` or name of
    the target actor which will receive the message.
:parameter action: the name of the :ref:`remote command <api-remote_commands>`
    to perform in the *target* :class:`Actor`.
:parameter args: positional arguments to pass to the
    :ref:`remote command <api-remote_commands>` *action*.
:parameter params: dictionary of parameters to pass to
    :ref:`remote command <api-remote_commands>` *action*.
:rtype: an :class:`ActorMessage` which is a :class:`Deferred` and therefore
    can be used to attach callbacks.

Typical example::

    >>> a = spawn()
    >>> r = a.add_callback(lambda p: send(p,'ping'))
    >>> r.result
    'pong'
'''
    return get_actor().send(target, action, *args, **params)


class Pulsar(LogginMixin):
    
    def configure_logging(self, **kwargs):
        configure_logging = super(Pulsar, self).configure_logging
        configure_logging(logger='pulsar',
                          config=self.cfg.logconfig,
                          level=self.cfg.loglevel,
                          handlers=self.cfg.loghandlers) 
        configure_logging(logger='pulsar.%s' % self.name,
                          config=self.cfg.logconfig,
                          level=self.cfg.loglevel,
                          handlers=self.cfg.loghandlers)
        

class Actor(Pulsar):
    '''The base class for concurrent programming in pulsar. In computer science,
the **Actor model** is a mathematical model of concurrent computation that
treats *actors* as the universal primitives of computation.
In response to a message that it receives, an actor can make local decisions,
create more actors, send more messages, and determine how to respond to
the next message received.

Pulsar actors are slightly different from the general theory. They cannot
create other actors, unless they are of special kind.

The current implementation allows for actors to perform specific tasks such
as listening to a socket, acting as http server, consuming
a task queue and so forth.

To spawn a new actor::

    >>> from pulsar import Actor, spawn
    >>> a = spawn(Actor)
    >>> a.is_alive()
    True

Here ``a`` is actually a reference to the remote actor.

**ATTRIBUTES**

.. attribute:: name

    The name of this :class:`Actor`.
    
.. attribute:: aid

    Unique ID for this :class:`Actor`.
    
.. attribute:: commands_set

    Set of :ref:`command names <actor_commands>` available to this
    :class:`Actor`.

.. attribute:: ioqueue

    An optional distributed queue. If present, it signal that the :class:`Actor`
    is a :ref:`CPU bound worker <cpubound>` receiving task requests on the
    :attr:`ioqueue`. 

    Default ``None``.

.. attribute:: cpubound

    Indicates if the :class:`Actor` is a :ref:`CPU-bound worker <cpubound>`
    or a I/O-bound one.
    CPU-bound actors have a separate event loop for handling I/O events.
    
    See also: :attr:`ioqueue`

.. attribute:: requestloop

    The :class:`IOLoop` to listen for requests.
    
.. attribute:: ioloop

    An instance of :class:`IOLoop` which listen for input/output events
    on a socket.  This is different from the :attr:`requestloop` only
    for :ref:`CPU-bound actors <cpubound>`.

.. attribute:: request_processed

    The total number of requests served by the actor

.. attribute:: concurrent_requests

    The current number of concurrent requests the actor is serving.
    Depending on the actor type, this number can be very high or max 1
    (CPU bound actors).
    
.. attribute:: proxy

    Instance of a :class:`ActorProxy` holding a reference
    to this :class:`Actor`. The proxy is a lightweight representation
    of the actor which can be shared across different processes
    (i.e. it is pickable).

.. attribute:: linked_actors

    Dictionary of :class:`ActorProxy` linked with this :class:`Actor`.

.. attribute:: params

    Contains parameters which are passed to actors spawned by this actor.
    
**METHODS**
'''
    exit_code = None
    mailbox = None
    signal_queue = None
    
    def __init__(self, impl):
        self.__impl = impl
        on_event = impl.params.pop('on_event', None)
        ioqueue = impl.params.pop('ioqueue', None)
        if not self.is_arbiter():
            if on_event:
                self.on_event = lambda fd, event: on_event(self, fd, event)
        else:
            ioqueue = None
        self.request_processed = 0
        self.concurrent_requests = 0
        self.state = ACTOR_STATES.INITIAL
        self.ioqueue = ioqueue
        self.linked_actors = impl.params.pop('linked_actors', {})
        self.monitors = impl.params.pop('monitors', {})
        self.arbiter = impl.params.pop('arbiter', None)
        self.monitor = impl.params.pop('monitor', None)
        self.proxy_mailboxes = {}
        self.params = AttributeDictionary(**impl.params)
        del impl.params
        setid(self)

    def __repr__(self):
        return self.impl.unique_name
    
    def __str__(self):
        return self.__repr__()
    
    ############################################################### PROPERTIES
    @property
    def impl(self):
        return self.__impl
    
    @property
    def aid(self):
        return self.__impl.aid
    
    @property
    def cfg(self):
        return self.__impl.cfg
    
    @property
    def name(self):
        return self.__impl.name
    
    @property
    def commands_set(self):
        return self.__impl.commands_set
    
    @property
    def proxy(self):
        return ActorProxy(self)

    @property
    def address(self):
        return self.mailbox.address

    @property
    def ioloop(self):
        return self.mailbox.ioloop

    @property
    def cpubound(self):
        if self.impl.kind != 'monitor':
            return self.ioqueue is not None
        else:
            return False

    @property
    def info_state(self):
        '''Current state description. One of ``initial``, ``running``,
 ``stopping``, ``closed`` and ``terminated``.'''
        return ACTOR_STATES.DESCRIPTION[self.state]

    ############################################################################
    ##    HIGH LEVEL API METHODS
    ############################################################################
    def start(self):
        '''Called after forking to start the actor's life. This is where
logging is configured, the :attr:`Actor.mailbox` is registered and the
:attr:`Actor.ioloop` is initialised and started.'''
        if self.state == ACTOR_STATES.INITIAL:
            self.configure_logging()
            self._setup_ioloop()
            events.fire('start', self)
            self.on_start()
            self.periodic_task()
            self.state = ACTOR_STATES.RUN
            self._run()
            
    def command(self, action):
        '''Fetch the pulsar command for *action*.'''
        return get_command(action, self.commands_set)

    def send(self, target, action, *args, **params):
        '''Send a message to *target* to perform *action* with given
parameters *params*. It return a :class:`ActorMessage`.'''
        if not isinstance(target, ActorProxy):
            target = get_proxy(self.get_actor(target))
        return target.receive_from(self, action, *args, **params)

    def put(self, request):
        '''Put a *request* into the :attr:`ioqueue` if available.'''
        if self.ioqueue:
            self.ioqueue.put(('request', request))
        else:
            self.logger.error("Trying to put a request on task queue,\
 but there isn't one!")
            
    ###############################################################  STATES
    def running(self):
        '''``True`` if actor is running.'''
        return self.state == ACTOR_STATES.RUN
    
    def active(self):
        '''``True`` if actor is active by being both running and having
the :attr:`ioloop` running.'''
        return self.running() and self.arbiter.aid in self.linked_actors
    
    def started(self):
        '''``True`` if actor has started. It does not necessarily
mean it is running.'''
        return self.state >= ACTOR_STATES.RUN

    def closed(self):
        '''``True`` if actor has exited in an clean fashion.'''
        return self.state == ACTOR_STATES.CLOSE

    def stopped(self):
        '''``True`` if actor has exited.'''
        return self.state >= ACTOR_STATES.CLOSE

    def is_arbiter(self):
        '''Return ``True`` if ``self`` is the :class:`Arbiter`.'''
        return False

    def is_monitor(self):
        '''Return ``True`` if ``self`` is a :class:`Monitor`.'''
        return False

    def is_process(self):
        '''boolean indicating if this is an actor on a child process.'''
        return self.impl.kind == 'process'

    def can_poll(self):
        '''Check if the actor can poll requests. This is used by 
:ref:`CPU-bound actors <cpubound>` only.'''
        if self.active():
            m = self.cfg.backlog
            return self.concurrent_requests < m if m else True
        else:
            return False

    def __reduce__(self):
        raise pickle.PicklingError('{0} - Cannot pickle Actor instances'\
                                   .format(self))

    ############################################################################
    ##    EVENT HANDLING
    ############################################################################
    @async()
    def handle_fd_event(self, fd, event):
        '''This function should be used when registering events
 on file descriptors registered with the :attr:`requestloop`.'''
        self.request_processed += 1
        self.concurrent_requests += 1
        try:
            future = make_async(self.on_event(fd, event))
        except Exception as e:
            result = as_failure(e)
        else:
            yield future
            result = future.result
        self.concurrent_requests -= 1
        log_failure(result)
        max_requests = self.cfg.max_requests
        if max_requests and self.request_processed >= max_requests:
            self.logger.warn("Shutting down %s. Max requests reached.", self)
            self.stop()

    ############################################################################
    ##    CALLBACKS
    ############################################################################
    def on_start(self):
        '''The :ref:`actor callback <actor-callbacks>` run **once** just before
the actor starts (after forking) its event loop. Every attribute is available,
therefore this is a chance to setup to perform custom initialisation
before the actor starts running.'''
        pass

    def on_event(self, fd, event):
        '''Handle an event on a file descriptor *fd*. This is what defines the
life of an actor.'''
        pass

    def on_stop(self):
        '''The :ref:`actor callback <actor-callbacks>` run once just before
 the actor stops running.'''
        pass

    def on_exit(self):
        '''The :ref:`actor callback <actor-callbacks>` run once when the actor
 has stopped running, just before it vanish in the garbage collector.'''
        pass

    def on_info(self, data):
        '''An :ref:`actor callback <actor-callbacks>` executed when
 obtaining information about the actor. It can be used to add additional
 data to the *data* dictionary. Information about the actor is obtained
 via the :meth:`Actor.info` method which is also exposed
 as a remote function.

 :parameter data: dictionary of data with information about the actor.
 :rtype: a dictionary of pickable data.'''
        return data

    ############################################################################
    # STOPPING
    ############################################################################
    def stop(self, force=False, exit_code=None):
        '''Stop the actor by stopping its :attr:`Actor.requestloop`
and closing its :attr:`Actor.mailbox`. Once everything is closed
properly this actor will go out of scope.'''
        if force or self.state <= ACTOR_STATES.RUN:
            events.fire('stop', self)
            self.state = ACTOR_STATES.STOPPING
            self.exit_code = exit_code
            try:
                res = self.on_stop()
            except:
                self.logger.error('Unhandle error while stopping',
                                  exc_info=True)
                res = None
            return res.addBoth(self.exit) if is_async(res) else self.exit()
    
    def exit(self, result=None):
        '''Exit from the :class:`Actor` domain.'''
        if not self.stopped():
            # we don't want monitors to stop the request loop
            if not self.is_monitor():
                self.requestloop.stop()
            self.mailbox.close()
            self.state = ACTOR_STATES.CLOSE
            self.logger.debug('%s exited', self)
            remove_actor(self)
            events.fire('exit', self)
            self.on_exit()

    ############################################################################
    #    INTERNALS
    ############################################################################
    def periodic_task(self):
        if self.can_continue():
            if self.active():
                self.send('arbiter', 'notify', self.info())
                secs = max(ACTOR_TIMEOUT_TOLERANCE*self.cfg.timeout, MIN_NOTIFY)
                next = time() + min(secs, MAX_NOTIFY)
                self.ioloop.add_timeout(next, self.periodic_task)
            else:
                self.ioloop.add_callback(self.periodic_task, False)
        
    def proxy_mailbox(self, address):
        m = self.proxy_mailboxes.get(address)
        if not m:
            try:
                m = mailbox(address=address)
            except socket.error:
                m = None
            else:
                self.proxy_mailboxes[address] = m
        return m
    
    def get_actor(self, aid):
        '''Given an actor unique id return the actor proxy.'''
        if aid == self.aid:
            return self
        elif aid == 'arbiter':
            return self.arbiter or self
        elif self.arbiter and aid == self.arbiter.aid:
            return self.arbiter
        elif self.monitor and aid == self.monitor.aid:
            return self.monitor
        elif aid in self.linked_actors:
            return self.linked_actors[aid]
        else:
            return self.monitors.get(aid)

    def info(self):
        '''return A dictionary of information related to the actor
status and performance.'''
        if not self.started():
            return
        isp = self.is_process()
        requestloop  = self.requestloop
        actor = {'name': self.name,
                 'state': self.info_state,
                 'actor_id': self.aid,
                 'uptime': time() - requestloop._started,
                 'thread_id': self.tid,
                 'process_id': self.pid,
                 'is_process': isp,
                 'internal_connections': self.mailbox.active_connections,
                 'age': self.impl.age}
        events = {'request processed': self.request_processed,
                  'callbacks': len(self.ioloop._callbacks),
                  'io_loops': self.ioloop.num_loops}
        if self.cpubound:
            events['request_loops'] = requestloop.num_loops
        data = {'actor': actor, 'events': events}
        if isp:
            data['system'] = system.system_info(self.pid)
        return self.on_info(data)
    
    def link_actor(self, proxy, address=None):
        '''Add the *proxy* to the :attr:`linked_actors` dictionary.
if *proxy* is not a class:`ActorProxy` instance raise an exception.'''
        if address:
            proxy.address = address
        if not proxy.address:
            raise ValueError('Linking with a actor without address')
        self.linked_actors[proxy.aid] = proxy
        # If the proxy is the actor monitor, add the arbiter
        # if the monitor is not the arbiter itself.
        # This last check is crucial in order to recursive call
        # causing stack overflow!
        if self.monitor == proxy and self.monitor != self.arbiter:
            self.link_actor(self.arbiter, address)
        return proxy

    def _setup_ioloop(self):
        # Internal function called at the start of the actor. It builds the
        # event loop which will consume events on file descriptors
        # Inject self as the actor of this thread
        ioq = self.ioqueue
        self.requestloop = IOLoop(io=IOQueue(ioq, self) if ioq else None,
                                  poll_timeout=self.params.poll_timeout,
                                  logger=self.logger)
        # If CPU bound add the request handler to the request loop
        if self.cpubound:
            self.requestloop.add_handler('request',
                                          self.handle_fd_event,
                                          self.requestloop.READ)
        if self.is_process():
            random.seed()
            proc_name = "%s-%s" % (self.cfg.proc_name, self)
            if system.set_proctitle(proc_name):
                self.logger.debug('Set process title to %s', proc_name)
            #system.set_owner_process(cfg.uid, cfg.gid)
            if is_mainthread():
                self.logger.debug('Installing signals')
                self.signal_queue = ThreadQueue()
                for name in system.ALL_SIGNALS:
                    sig = getattr(signal, 'SIG%s' % name)
                    try:
                        signal.signal(sig, self._queue_signal)
                    except ValueError:
                        break
        self.mailbox = mailbox(self)
        set_local_data(self)
        self.logger.info('%s started at address %s', self, self.mailbox.address)

    def _queue_signal(self, sig, frame=None):
        self.signal_queue.put(sig)
        
    def can_continue(self):
        if self.signal_queue is not None:
            while True:
                try:
                    sig = self.signal_queue.get(timeout=0.01)
                except (Empty, IOError):
                    break
                if sig not in system.SIG_NAMES:
                    self.logger.info("Ignoring unknown signal: %s", sig)
                else:
                    signame = system.SIG_NAMES.get(sig)
                    if sig in EXIT_SIGNALS:
                        self.logger.warn("Got signal %s. Stopping.", signame)
                        self.stop(True)
                        return False
                    else:
                        self.logger.debug('No handler for signal %s.', signame)
        return True
    
    def _run(self):
        try:
            self.requestloop.start()
        finally:
            self.stop()



import zlib
from libc.stdlib cimport *
from cpython cimport PyBytes_FromStringAndSize

    
cdef inline bytes_to_str(b):
    try:
        return str(b, 'latin1')
    except TypeError:
        return b
        

cdef extern from "http_parser.h" nogil:
    
    cdef enum http_method:
        HTTP_DELETE, HTTP_GET, HTTP_HEAD, HTTP_POST, HTTP_PUT,
        HTTP_CONNECT, HTTP_OPTIONS, HTTP_TRACE, HTTP_COPY, HTTP_LOCK,
        HTTP_MKCOL, HTTP_MOVE, HTTP_PROPFIND, HTTP_PROPPATCH, HTTP_UNLOCK, 
        HTTP_REPORT, HTTP_MKACTIVITY, HTTP_CHECKOUT, HTTP_MERGE, HTTP_MSEARCH,
        HTTP_NOTIFY, HTTP_SUBSCRIBE, HTTP_UNSUBSCRIBE, HTTP_PATCH
    
        
    cdef enum http_parser_type:
        HTTP_REQUEST, HTTP_RESPONSE, HTTP_BOTH
    
        
    cdef struct http_parser:
        int content_length
        unsigned short http_major
        unsigned short http_minor
        unsigned short status_code
        unsigned char method
        char upgrade
        void *data


    ctypedef int (*http_data_cb) (http_parser*, char *at, size_t length)
    ctypedef int (*http_cb) (http_parser*)
    
            
    struct http_parser_settings:
        http_cb on_message_begin
        http_data_cb on_url
        http_data_cb on_header_field
        http_data_cb on_header_value
        http_cb on_headers_complete
        http_data_cb on_body
        http_cb on_message_complete


    void http_parser_init(http_parser *parser, 
            http_parser_type ptype)
    
    
    size_t http_parser_execute(http_parser *parser, 
            http_parser_settings *settings, char *data,
            size_t len)


    int http_should_keep_alive(http_parser *parser)


    char *http_method_str(http_method)


cdef inline int on_url_cb(http_parser *parser, char *at,
                          size_t length):
    res = <object>parser.data
    value = bytes_to_str(PyBytes_FromStringAndSize(at, length))
    res.url = value
    res.environ['RAW_URI'] = value
    return 0


cdef inline int on_header_field_cb(http_parser *parser, char *at, 
        size_t length):
    header_field = PyBytes_FromStringAndSize(at, length)
    res = <object>parser.data
    if res._last_was_value:
        res._last_field = ""
    res._last_field += bytes_to_str(header_field)
    res._last_was_value = False
    return 0


cdef inline int on_header_value_cb(http_parser *parser, char *at, 
                                   size_t length):
    res = <object>parser.data
    header_value = bytes_to_str(PyBytes_FromStringAndSize(at, length))
    if res._last_field in res.headers:
        header_value = "%s, %s" % (res.headers[res._last_field],
                header_value)
    res.headers[res._last_field] = header_value
    res._last_was_value = True
    return 0


cdef inline int on_headers_complete_cb(http_parser *parser):
    res = <object>parser.data
    res.headers_complete = True

    if res.decompress:
        encoding = res.headers.get('content-encoding')
        if encoding == 'gzip':
            res.decompressobj = zlib.decompressobj(16+zlib.MAX_WBITS)
            del res.headers['content-encoding']
        elif encoding == 'deflate':
            res.decompressobj = zlib.decompressobj()
            del res.headers['content-encoding']
        else:
            res.decompress = False
    
    return 0


cdef inline int on_message_begin_cb(http_parser *parser):
    res = <object>parser.data
    res.message_begin = True
    return 0


cdef inline int on_body_cb(http_parser *parser, char *at, 
        size_t length):
    res = <object>parser.data
    value = PyBytes_FromStringAndSize(at, length)

    res.partial_body = True

    # decompress the value if needed
    if res.decompress:
        value = res.decompressobj.decompress(value)

    res.body.append(value)
    return 0


cdef inline int on_message_complete_cb(http_parser *parser):
    res = <object>parser.data
    res.message_complete = True
    return 0
    



import sys
import io
import time
import logging
import socket
import errno
from collections import deque
from threading import current_thread, Thread

from pulsar.utils.system import IObase
from pulsar import create_socket, server_socket, create_client_socket,\
                     wrap_socket, defaults, create_connection, CouldNotParse,\
                     get_socket_timeout, Timeout, BaseSocket, Synchronized
from pulsar.utils.httpurl import IOClientRead
from pulsar.utils.structures import merge_prefix
from .defer import Deferred, is_async, is_failure, async, maybe_async,\
                        safe_async, log_failure, NOT_DONE, range
from .eventloop import IOLoop, loop_timeout
from .access import PulsarThread, thread_ioloop, get_actor

LOGGER = logging.getLogger('pulsar.iostream')


__all__ = ['AsyncIOStream',
           'ProtocolSocket',
           'ClientSocket',
           'Client',
           'ClientSocketHandler',
           'AsyncConnection',
           'AsyncResponse',
           'AsyncSocketServer',
           'WRITE_BUFFER_MAX_SIZE']

WRITE_BUFFER_MAX_SIZE = 128 * 1024  # 128 kb

ASYNC_ERRNO = (errno.EWOULDBLOCK, errno.EAGAIN, errno.EINPROGRESS)
def async_error(e):
    return e.args and e.args[0] in ASYNC_ERRNO
    

class AsyncIOStream(IObase, BaseSocket):
    ''':ref:`Framework class <pulsar_framework>` to write and read
from a non-blocking socket. It is used everywhere in :mod:`pulsar` for
handling asynchronous :meth:`write` and :meth:`read` operations with
`callbacks` which can be used to act when data has just been sent or has
just been received.

It was originally forked from tornado_ IOStream and subsequently
adapted to pulsar :ref:`concurrent framework <design>`.

.. attribute:: socket

    A :class:`Socket` which might be connected or unconnected.

.. attribute:: timeout

    A timeout in second which is used when waiting for a
    data to be available for reading. If timeout is a positive number,
    every time the :class:`AsyncIOStream` performs a :meth:`read`
    operation a timeout is also created on the :attr:`ioloop`.
'''
    _error = None
    _socket = None
    _state = None
    _read_timeout = None
    _read_callback = None
    _read_length = None
    _write_callback = None
    _close_callback = None
    _connect_callback = None

    def __init__(self, socket=None, max_buffer_size=None,
                 read_chunk_size=None, timeout=None):
        self.sock = socket
        self._read_callback_timeout = timeout
        self.max_buffer_size = max_buffer_size or 104857600
        self.read_chunk_size = read_chunk_size or io.DEFAULT_BUFFER_SIZE
        self._read_buffer = deque()
        self._write_buffer = deque()

    def __repr__(self):
        if self.sock:
            return '%s (%s)' % (self.sock, self.state_code)
        else:
            return '(closed)'

    def __str__(self):
        return self.__repr__()

    #######################################################    STATES
    @property
    def connecting(self):
        return self._connect_callback is not None

    @property
    def reading(self):
        """Returns true if we are currently reading from the stream."""
        return self._read_callback is not None

    @property
    def writing(self):
        """Returns true if we are currently writing to the stream."""
        return bool(self._write_buffer)

    @property
    def closed(self):
        '''Boolean indicating if the :attr:`sock` is closed.'''
        return self.sock is None

    @property
    def state(self):
        return self._state
    
    @property
    def error(self):
        return self._error

    @property
    def state_code(self):
        s = []
        if self.closed:
            return 'closed'
        if self.connecting:
            s.append('connecting')
        if self.writing:
            s.append('writing')
        if self.reading:
            s.append('reading')
        return ' '.join(s) if s else 'idle'

    @property
    def ioloop(self):
        return thread_ioloop()
    
    def settimeout(self, value):
        pass

    def _set_socket(self, sock):
        if self._socket is None:
            self._socket = wrap_socket(sock)
            self._state = None
            if self._socket is not None:
                self._socket.settimeout(0)
        else:
            raise RuntimeError('Cannot set socket. Close the existing one.')
    def _get_socket(self):
        return self._socket
    sock = property(_get_socket, _set_socket)

    #######################################################    ACTIONS
    def connect(self, address):
        """Connects the socket to a remote address without blocking.
May only be called if the socket passed to the constructor was not available
or it was not previously connected.  The address parameter is in the
same format as for socket.connect, i.e. a (host, port) tuple or a string
for unix sockets.
If callback is specified, it will be called when the connection is completed.
Note that it is safe to call IOStream.write while the
connection is pending, in which case the data will be written
as soon as the connection is ready.  Calling IOStream read
methods before the socket is connected works on some platforms
but is non-portable."""
        if self._state is None and not self.connecting:
            if self.sock is None:
                self.sock = create_client_socket(address)
            try:
                self.sock.connect(address)
            except socket.error as e:
                # In non-blocking mode connect() always raises an exception
                if not async_error(e):
                    LOGGER.warning('Connect error on %s: %s', self, e)
                    self.close()
                    return
            callback = Deferred(description='%s connect callback' % self)
            self._connect_callback = callback
            self._add_io_state(self.WRITE)
            return callback
        else:
            raise RuntimeError('Cannot connect. State is %s.' % self.state_code)

    def read(self, length=None):
        """Starts reading data from the :attr:`sock`. It returns a
:class:`Deferred` which will be called back once data is available.
If this function is called while this class:`AsyncIOStream` is already reading
a RuntimeError occurs.

:rtype: a :class:`pulsar.Deferred` instance.

One common pattern of usage::

    def parse(data):
        ...

    io = AsyncIOStream(socket=sock)
    io.read().add_callback(parse)

"""
        if self.reading:
            raise RuntimeError("Asynchronous stream %s already reading!" %
                               str(self.address))
        if self.closed:
            return self._get_buffer(self._read_buffer)
        else:
            callback = Deferred(description='%s read callback' % self)
            self._read_callback = callback
            self._read_length = length
            self._add_io_state(self.READ)
            if self._read_timeout:
                try:
                    self.ioloop.remove_timeout(self._read_timeout)
                except ValueError:
                    pass
            self._read_timeout = loop_timeout(callback,
                                              self._read_callback_timeout,
                                              self.ioloop)
            return callback
    recv = read

    def write(self, data):
        """Write the given *data* to this stream. If there was previously
buffered write data and an old write callback, that callback is simply
overwritten with this new callback.

:rtype: a :class:`Deferred` instance or the number of bytes written.
        """
        self._check_closed()
        if data:
            assert isinstance(data, bytes)
            if len(data) > WRITE_BUFFER_MAX_SIZE:
                for i in range(0, len(data), WRITE_BUFFER_MAX_SIZE):
                    self._write_buffer.append(data[i:i+WRITE_BUFFER_MAX_SIZE])
            else:
                self._write_buffer.append(data)
        #
        if not self.connecting:
            tot_bytes = self._handle_write()
            # data still in the buffer
            if self._write_buffer:
                callback = Deferred(description='%s write callback' % self)
                self._write_callback = callback
                self._add_io_state(self.WRITE)
                return callback
            else:
                return tot_bytes
    sendall = write

    def close(self):
        """Close the :attr:`sock` and call the *callback* if it was
setup using the :meth:`set_close_callback` method."""
        if not self.closed:
            exc_info = sys.exc_info()
            if any(exc_info):
                self._error = exc_info[1]
            if self._state is not None:
                self.ioloop.remove_handler(self.fileno())
            self.sock.close()
            self._socket = None
            if self._close_callback:
                self._may_run_callback(self._close_callback)

    def set_close_callback(self, callback):
        """Call the given callback when the stream is closed."""
        self._close_callback = callback

    #######################################################    INTERNALS
    def read_to_buffer(self):
        #Reads from the socket and appends the result to the read buffer.
        #Returns the number of bytes read.
        length = self._read_length or self.read_chunk_size
        # Read from the socket until we get EWOULDBLOCK or equivalent.
        # SSL sockets do some internal buffering, and if the data is
        # sitting in the SSL object's buffer select() and friends
        # can't see it; the only way to find out if it's there is to
        # try to read it.
        while True:
            try:
                chunk = self.sock.recv(length)
            except socket.error as e:
                if async_error(e):
                    chunk = None
                else:
                    raise
            if not chunk:
                # if chunk is b'' close the socket
                if chunk is not None:
                    self.close()
                break
            self._read_buffer.append(chunk)
            if self._read_buffer_size() >= self.max_buffer_size:
                LOGGER.error("Reached maximum read buffer size")
                self.close()
                raise IOError("Reached maximum read buffer size")
            if len(chunk) < length:
                break
        return self._read_buffer_size()

    def _read_buffer_size(self):
        return sum(len(chunk) for chunk in self._read_buffer)

    def _may_run_callback(self, c, result=None):
        try:
            # Make sure that any uncaught error is logged
            log_failure(c.callback(result))
        except:
            # Close the socket on an uncaught exception from a user callback
            # (It would eventually get closed when the socket object is
            # gc'd, but we don't want to rely on gc happening before we
            # run out of file descriptors)
            self.close()

    def _handle_connect(self):
        callback = self._connect_callback
        self._connect_callback = None
        self._may_run_callback(callback)

    def _handle_read(self):
        try:
            try:
                result = self.read_to_buffer()
            except (socket.error, IOError, OSError) as e:
                if e.args and e.args[0] == errno.ECONNRESET:
                    # Treat ECONNRESET as a connection close rather than
                    # an error to minimize log spam  (the exception will
                    # be available on self.error for apps that care).
                    result = 0
                else:
                    raise
        except Exception:
            result = 0
            LOGGER.warning("Read error on %s.", self, exc_info=True)
        if result == 0:
            self.close()
        buffer = self._get_buffer(self._read_buffer)
        if self.reading:
            callback = self._read_callback
            self._read_callback = None
            self._may_run_callback(callback, buffer)

    def _handle_write(self):
        # keep count how many bytes we write
        tot_bytes = 0
        while self._write_buffer:
            try:
                sent = self.sock.send(self._write_buffer[0])
                if sent == 0:
                    # With OpenSSL, after send returns EWOULDBLOCK,
                    # the very same string object must be used on the
                    # next call to send.  Therefore we suppress
                    # merging the write buffer after an EWOULDBLOCK.
                    # A cleaner solution would be to set
                    # SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER, but this is
                    # not yet accessible from python
                    # (http://bugs.python.org/issue8240)
                    break
                merge_prefix(self._write_buffer, sent)
                self._write_buffer.popleft()
                tot_bytes += sent
            except socket.error as e:
                if async_error(e):
                    break
                else:
                    LOGGER.warning("Write error on %s: %s", self, e)
                    self.close()
                    return
        if not self._write_buffer and self._write_callback:
            callback = self._write_callback
            self._write_callback = None
            self._may_run_callback(callback, tot_bytes)
        return tot_bytes

    def _check_closed(self):
        if not self.sock:
            raise IOError("Stream is closed")

    def _get_buffer(self, dq):
        buff = b''.join(dq)
        dq.clear()
        return buff

    def _handle_events(self, fd, events):
        # This is the actual callback from the event loop
        if not self.sock:
            LOGGER.warning("Got events for closed stream %d", fd)
            return
        try:
            if events & self.READ:
                self._handle_read()
            if not self.sock:
                return
            if events & self.WRITE:
                if self.connecting:
                    self._handle_connect()
                self._handle_write()
            if not self.sock:
                return
            if events & self.ERROR:
                # We may have queued up a user callback in _handle_read or
                # _handle_write, so don't close the IOStream until those
                # callbacks have had a chance to run.
                self.ioloop.add_callback(self.close)
                return
            state = self.ERROR
            if self.reading:
                state |= self.READ
            if self.writing:
                state |= self.WRITE
            if state != self._state:
                assert self._state is not None, \
                    "shouldn't happen: _handle_events without self._state"
                self._state = state
                self.ioloop.update_handler(self.fileno(), self._state)
        except:
            self.close()
            raise

    def _add_io_state(self, state):
        if self.sock is None:
            # connection has been closed, so there can be no future events
            return
        if self._state is None:
            # If the state was not set add the handler to the event loop
            self._state = self.ERROR | state
            self.ioloop.add_handler(self, self._handle_events, self._state)
        elif not self._state & state:
            # update the handler
            self._state = self._state | state
            self.ioloop.update_handler(self, self._state)


class EchoParser:
    '''A simple echo protocol'''
    def encode(self, data):
        return data
    
    def decode(self, data):
        return bytes(data), bytearray()
    
    
class ProtocolSocket(BaseSocket):
    '''A :class:`BaseSocket` with a protocol for encoding and decoding
messages. This is the base class for :class:`AsyncSocketServer`,
:class:`AsyncConnection` and :class:`ClientSocketHandler`.

.. attribute:: on_closed

    A :class:`Deferred` which receives a callback once the
    :meth:`close` method is invoked.
 
'''
    protocol_factory = EchoParser
    '''A callable for building the protocol to encode and decode messages. This
    attribute can be specified as class attribute or via the constructor.
    A :ref:`protocol <socket-protocol>` must be an instance of a class
    exposing the ``encode`` and ``decode`` methods.'''
    _closing_socket = False
    def __new__(cls, *args, **kwargs):
        o = super(ProtocolSocket, cls).__new__(cls)
        o.on_closed = Deferred(
                        description='on_closed %s callback' % cls.__name__)
        o.time_started = time.time()
        o.time_last = o.time_started
        f = kwargs.get('protocol_factory')
        if f:
            o.protocol_factory = f
        o.received = 0
        return o

    @property
    def closed(self):
        '''True if the socket is closed.'''
        return self.sock.closed

    def on_close(self, failure=None):
        '''Callback just before closing the socket'''
        pass

    def close(self, msg=None):
        '''Close this socket and log the failure if there was one.'''
        if self._closing_socket:
            return msg
        self._closing_socket = True
        if is_failure(msg):
            if isinstance(msg.trace[1], Timeout):
                if not msg.logged:
                    msg.logged = True
                    LOGGER.info('Closing %s on timeout.', self)
            else:
                log_failure(msg)
        self.on_close(msg)
        self.sock.close()
        return self.on_closed.callback(msg)
    
    
class ClientSocketHandler(ProtocolSocket):
    '''A :class:`ProtocolSocket` for clients.
This class can be used with synchronous and asynchronous socket
for both a "client" socket and a server connection socket (the socket
obtained from a server socket via the ``connect`` function).

.. attribute:: protocol

    An object obtained by the :attr:`ProtocolSocket.protocol_factory` method.
    It must implement the ``encode`` and ``decode`` methods used, respectively,
    by clients and servers.
    
.. attribute:: buffer

    A bytearray containing data received from the socket. This buffer
    is not empty when more data is required to form a message.
    
.. attribute:: remote_address

    The remote address for the socket communicating with this
    :class:`ClientSocketHandler`.
'''
    def __init__(self, socket, address, timeout=None, read_timeout=None,
                 **kwargs):
        '''Create a client or client-connection socket.

:parameter socket: a client or client-connection socket
:parameter address: The address of the remote client/server
:parameter protocol_factory: A callable used for creating
    :attr:`ProtoSocket.protocol` instance.
:parameter timeout: A timeout in seconds for the socket. Same rules as
    the ``socket.settimeout`` method in the standard library. A value of 0
    indicates an asynchronous socket.
:parameter read_timeout: A timeout in seconds for asynchronous operations. This
    value is only used when *timeout* is 0 in the constructor
    of a :class:`AsyncIOStream`.
'''
        self._socket_timeout = get_socket_timeout(timeout)
        self._set_socket(socket, read_timeout)
        self.remote_address = address
        self.protocol = self.protocol_factory()
        self.buffer = bytearray()

    def __repr__(self):
        return str(self.remote_address)
    __str__ = __repr__

    def _set_socket(self, sock, read_timeout=None):
        if not isinstance(sock, AsyncIOStream):
            if self._socket_timeout == 0:
                sock = AsyncIOStream(sock)
            else:
                sock = wrap_socket(sock)
                sock.settimeout(self._socket_timeout)
        self.sock = sock
        if self.async:
            close_callback = Deferred().add_callback(self.close)
            self.sock.set_close_callback(close_callback)
            self.read_timeout = read_timeout
            
    def _get_read_timeout(self):
        if self.async:
            return self.sock._read_callback_timeout
        else:
            return self._socket_timeout
    def _set_read_timeout(self, value):
        if self.async:
            self.sock._read_callback_timeout = value
        else:
            self._socket_timeout = value
            self.sock.settimeout(self._socket_timeout)
    read_timeout = property(_get_read_timeout, _set_read_timeout) 


class ClientSocket(ClientSocketHandler, IOClientRead):
    '''Synchronous/Asynchronous client for a remote socket server. This client
maintain a connection with remote server and exchange data by writing and
reading from the same socket connection.'''
    def __init__(self, *args, **kwargs):
        super(ClientSocket, self).__init__(*args, **kwargs)
        self.exec_queue = deque()
        self.processing = False
        
    @classmethod
    def connect(cls, address, **kwargs):
        '''Create a new :class:`ClientSocket` connected at *address*.'''
        sock = create_connection(address)
        return cls(sock, address, **kwargs)

    def send(self, data):
        '''Send data to remote server'''
        self.time_last = time.time()
        data = self.protocol.encode(data)
        return self.sock.write(data)
        
    def execute(self, data):
        '''Send and read data from socket. It makes sure commands are queued
and executed in an orderly fashion. For asynchronous connection it returns
a :class:`Deferred` called once data has been received from the server and
parsed.'''
        if self.async:
            cbk = Deferred()
            if data:
                cbk.addBoth(self._got_result)
                self.exec_queue.append((data, cbk))
                self.sock.ioloop.add_callback(self._consume_next)
            else:
                cbk.callback(None)
            return cbk
        elif data:
            self.send(data)
            return self._read()
        
    def parsedata(self, data):
        '''We got some data to parse'''
        parsed_data = self._parsedata(data)
        if parsed_data:
            self.received += 1
            return parsed_data
        
    def _consume_next(self):
        # This function is always called in the socket IOloop
        if not self.processing and self.exec_queue:
            self.processing = True
            data, cbk = self.exec_queue.popleft()
            msg = safe_async(self.send, (data,))\
                    .add_callback(self._read, self.close)
            msg = maybe_async(msg)
            if is_async(msg):
                msg.addBoth(cbk.callback)
            else:
                cbk.callback(msg)
            
    def _got_result(self, result):
        # This callback is always executed in the socket IOloop
        self.processing = False
        # keep on consuming if needed
        self._consume_next()
        return result

    def _parsedata(self, data):
        buffer = self.buffer
        if data:
            # extend the buffer
            buffer.extend(data)
        elif not buffer:
            return
        try:
            parsed_data, buffer = self.protocol.decode(buffer)
        except CouldNotParse:
            LOGGER.warn('Could not parse data', exc_info=True)
            parsed_data = None
            buffer = bytearray()
        self.buffer = buffer
        return parsed_data


class Client(ClientSocket):

    def reconnect(self):
        if self.closed:
            sock = create_connection(self.remote_address, blocking=True)
            self._set_socket(sock)


class ReconnectingClient(Client):

    def send(self, data):
        self.reconnect()
        return super(ReconnectingClient, self).send(data)


class AsyncResponse(object):
    '''An asynchronous server response is created once an
:class:`AsyncConnection` has available parsed data from a read operation.
Instances of this class are iterable and produce chunk of data to send back
to the remote client.

The ``__iter__`` is the only method which **needs** to be implemented by
derived classes. If an empty byte is yielded, the asynchronous engine
will resume the iteration after one loop in the actor event loop.

.. attribute:: connection

    The :class:`AsyncConnection` for this response
    
.. attribute:: server

    The :class:`AsyncSocketServer` for this response

.. attribute:: parsed_data

    Parsed data from remote client
'''
    def __init__(self, connection, parsed_data):
        self.connection = connection
        self.parsed_data = parsed_data

    @property
    def server(self):
        return self.connection.server

    @property
    def protocol(self):
        return self.connection.protocol

    @property
    def sock(self):
        return self.connection.sock

    def __iter__(self):
        # by default it echos the client message
        yield self.protocol.encode(self.parsed_data)


class AsyncConnection(ClientSocketHandler):
    '''An asynchronous client connection for a :class:`AsyncSocketServer`.
The connection maintains the client socket open for as long as it is required.
A connection can handle several request/responses until it is closed.

.. attribute:: server

    The class :class:`AsyncSocketServer` which created this
    :class:`AsyncConnection`.

.. attribute:: response_class

    Class or callable for building an :class:`AsyncResponse` object. It is
    initialised by the :class:`AsyncSocketServer.response_class` but it can be
    changed at runtime when upgrading connections to new protocols. An example
    is the websocket protocol.
'''
    def __init__(self, sock, address, server, **kwargs):
        if not isinstance(sock, AsyncIOStream):
            sock = AsyncIOStream(sock, timeout=server.timeout)
        super(AsyncConnection, self).__init__(sock, address, **kwargs)
        self.server = server
        self.response_class = server.response_class
        server.connections.add(self)
        self.handle().add_errback(self.close)

    @async(max_errors=1, description='Asynchronous client connection generator')
    def handle(self):
        while not self.closed:
            # Read the socket
            outcome = self.sock.read()
            yield outcome
            # if we are here it means no errors occurred so far and
            # data is available to process (max_errors=1)
            self.time_last = time.time()
            buffer = self.buffer
            buffer.extend(outcome.result)
            parsed_data = True
            # IMPORTANT! Consume all data until the protocol returns nothing.
            while parsed_data:
                parsed_data = self.request_data()
                if parsed_data:
                    self.received += 1
                    # The response is an iterable (same as WSGI response)
                    for data in self.response_class(self, parsed_data):
                        if data:
                            yield self.write(data)
                        else: # The response is not ready. release the loop
                            yield NOT_DONE
    
    def request(self, response=None):
        if self._current_request is None:
            self._current_request = AsyncRequest(self)
        request = self._current_request
        if response is not None:
            self._current_request = None
            request.callback(response)
        return request

    def request_data(self):
        '''This function is called when data to parse is available on the
:attr:`ClientSocket.buffer`. It should return parsed data or ``None`` if
more data in the buffer is required.'''
        buffer = self.buffer
        if not buffer:
            return
        self.protocol.connection = self
        try:
            parsed_data, buffer = self.protocol.decode(buffer)
        except:
            LOGGER.error('Could not parse data', exc_info=True)
            raise
        self.buffer = buffer
        return parsed_data

    @property
    def actor(self):
        return self.server.actor

    def on_close(self, failure=None):
        self.server.connections.discard(self)

    def write(self, data):
        '''Write data to socket.'''
        return self.sock.write(data)


class AsyncSocketServer(ProtocolSocket):
    '''A :class:`ProtocolSocket` for asynchronous servers which listen
for requests on a socket.

.. attribute:: actor

    The :class:`Actor` running this :class:`AsyncSocketServer`.

.. attribute:: ioloop

    The :class:`IOLoop` used by this :class:`AsyncSocketServer` for
    asynchronously sending and receiving data.

.. attribute:: connections

    The set of all open :class:`AsyncConnection`

.. attribute:: onthread

    If ``True`` the server has its own :class:`IOLoop` running on a separate
    thread of execution. Otherwise it shares the :attr:`actor.requestloop`

.. attribute:: timeout

    The timeout when reading data in an asynchronous way.
    
'''
    thread = None
    _started = False
    connection_class = AsyncConnection
    '''A subclass of :class:`AsyncConnection`. A new instance of this class is
constructued each time a new connection has been established by the
:meth:`accept` method.'''
    response_class = AsyncResponse
    '''A subclass of :class:`AsyncResponse` for handling responses to clients
    once data has been received and processed.'''
    
    def __init__(self, actor, socket, onthread=False, connection_class=None,
                 response_class=None, timeout=None, **kwargs):
        self.actor = actor
        self.connection_class = connection_class or self.connection_class
        self.response_class = response_class or self.response_class
        self.sock = wrap_socket(socket)
        self.connections = set()
        self.timeout = timeout
        self.on_connection_callbacks = []
        if onthread:
            # Create a pulsar thread and starts it
            self.__ioloop = IOLoop()
            self.thread = PulsarThread(name=self.name, target=self._run)
            self.thread.actor = actor
            actor.requestloop.add_callback(lambda: self.thread.start())
        else:
            self.__ioloop = actor.requestloop
            actor.requestloop.add_callback(self._run)
        
    @classmethod
    def make(cls, actor=None, bind=None, backlog=None, **kwargs):
        if actor is None:
            actor = get_actor()
        if not backlog:
            if actor:
                backlog = actor.cfg.backlog
            else:
                backlog = defaults.BACKLOG
        if bind:
            socket = create_socket(bind, backlog=backlog)
        else:
            socket = server_socket(backlog=backlog)
        return cls(actor, socket, **kwargs)

    @property
    def onthread(self):
        return self.thread is not None
    
    @property
    def name(self):
        return '%s %s' % (self.__class__.__name__, self.address)

    def __repr__(self):
        return self.name
    __str__ = __repr__

    @property
    def ioloop(self):
        return self.__ioloop

    @property
    def active_connections(self):
        return len(self.connections)

    def quit_connections(self):
        for c in list(self.connections):
            c.close()
            
    def on_close(self, failure=None):
        self.quit_connections()
        self.ioloop.remove_handler(self)
        if self.onthread:
            self.__ioloop.stop()
            # We join the thread
            if current_thread() != self.thread:
                try:
                    self.thread.join()
                except RuntimeError:
                    pass

    ############################################################## INTERNALS
    def _run(self):
        # add new_connection READ handler to the eventloop and starts the
        # eventloop if it was not already started.
        self.actor.logger.debug('Registering %s with event loop.', self)
        self.ioloop.add_handler(self,
                                self._on_connection,
                                self.ioloop.READ)
        self.ioloop.start()

    def _on_connection(self, fd, events):
        '''Called when a new connection is available.'''
        # obtain the client connection
        for callback in self.on_connection_callbacks:
            c = callback(fd, events)
            if c is not None:
                return c
        return self.accept()

    def accept(self):
        '''Accept a new connection from a remote client'''
        client, client_address = self.sock.accept()
        if client:
            self.received += 1
            return self.connection_class(client, client_address, self,
                                         protocol_factory=self.protocol_factory)



from time import time
import os
import sys
import signal
from multiprocessing import current_process

import pulsar
from pulsar.utils import system
from pulsar.utils.tools import Pidfile
from pulsar.utils.security import gen_unique_id
from pulsar import HaltServer

from .defer import itervalues, iteritems, multi_async, async
from .actor import Actor, ACTOR_STATES
from .monitor import PoolMixin, _spawn_actor
from .access import get_actor, set_actor
from . import proxy


process_global = pulsar.process_global

__all__ = ['arbiter', 'spawn', 'Arbiter']


def arbiter(commands_set=None, **params):
    '''Obtain the arbiter instance.'''
    arbiter = get_actor()
    if arbiter is None:
        # Create the arbiter
        cset = set(proxy.actor_commands)
        cset.update(proxy.arbiter_commands)
        cset.update(commands_set or ())
        return set_actor(_spawn_actor(Arbiter, commands_set=cset, **params))
    elif isinstance(arbiter, Actor) and arbiter.is_arbiter():
        return arbiter


def spawn(cfg=None, **kwargs):
    '''Spawn a new :class:`Actor` and return an :class:`ActorProxyDeferred`.
This method can be used from any :class:`Actor`.
If not in the :class:`Arbiter` domain,
the method send a request to the :class:`Arbiter` to spawn a new actor, once
the arbiter creates the actor it returns the proxy to the original caller.

**Parameter kwargs**

These optional parameters are:
    * *actor_class* a custom :class:`Actor` subclass.
    * *aid* the actor id
    * *commands_set* the set of :ref:`remote commands <api-remote_commands>`
      the :class:`Actor` can respond to.
    
:rtype: an :class:`ActorProxyDeferred`.

A typical usage::

    >>> a = spawn()
    >>> a.aid
    'ba42b02b'
    >>> a.called
    True
    >>> p = a.result
    >>> p.address
    ('127.0.0.1', 46691)
    '''
    aid = gen_unique_id()[:8]
    kwargs['aid'] = aid
    actor = get_actor()
    # The actor is not the Arbiter domain.
    # We send a message to the Arbiter to spawn a new Actor
    if not isinstance(actor, Arbiter):
        msg = actor.send('arbiter', 'spawn', **kwargs)\
                            .add_callback(actor.link_actor)
        return proxy.ActorProxyDeferred(aid, msg)
    else:
        return actor.spawn(**kwargs)

    
class Arbiter(PoolMixin, Actor):
    '''The Arbiter is the most important a :class:`Actor`
and :class:`PoolMixin` in pulsar concurrent framework. It is used as singleton
in the main process and it manages one or more :class:`Monitor`.
It runs the main :class:`IOLoop` of your concurrent application.
It is the equivalent of the gunicorn_ arbiter, the twisted_ reactor
and the tornado_ eventloop.

Users access the arbiter (in the arbiter process domain) by the high level api::

    import pulsar

    arbiter = pulsar.arbiter()

.. _gunicorn: http://gunicorn.org/
.. _twisted: http://twistedmatrix.com/trac/
.. _tornado: http://www.tornadoweb.org/
'''
    pidfile = None
    restarted = False

    ############################################################################
    # ARBITER HIGH LEVEL API
    ############################################################################
    def is_arbiter(self):
        return True

    def add_monitor(self, monitor_class, monitor_name, **params):
        '''Add a new :class:`Monitor` to the :class:`Arbiter`.

:parameter monitor_class: a :class:`pulsar.Monitor` class.
:parameter monitor_name: a unique name for the monitor.
:parameter kwargs: dictionary of key-valued parameters for the monitor.
:rtype: an instance of a :class:`pulsar.Monitor`.'''
        if monitor_name in self.monitors:
            raise KeyError('Monitor "{0}" already available'\
                           .format(monitor_name))
        params['name'] = monitor_name
        m = self.spawn(monitor_class, **params)
        self.linked_actors[m.aid] = m
        self.monitors[m.name] = m
        return m

    def is_process(self):
        return True

    def get_all_monitors(self):
        '''A dictionary of all :class:`Monitor` in the arbiter'''
        return dict(((mon.name, mon.proxy) for mon in\
                      itervalues(self.monitors) if mon.mailbox))

    @multi_async
    def close_monitors(self):
        '''Close all :class:`Monitor` at once.'''
        for pool in list(itervalues(self.monitors)):
            yield pool.stop(True)

    def on_info(self, data):
        monitors = [p.info() for p in itervalues(self.monitors)]
        server = data.pop('actor')
        server.update({'version': pulsar.__version__,
                       'name': pulsar.SERVER_NAME,
                       'number_of_monitors': len(self.monitors),
                       'number_of_actors': len(self.managed_actors)})
        server.pop('is_process', None)
        server.pop('ppid', None)
        server.pop('actor_id', None)
        server.pop('age', None)
        data['server'] = server
        data['workers'] = [a.info for a in itervalues(self.managed_actors)]
        data['monitors'] = monitors
        return data

    ############################################################################
    # OVERRIDE ACTOR HOOKS
    ############################################################################
    def on_start(self):
        if current_process().daemon:
            raise pulsar.PulsarException(
                    'Cannot create the arbiter in a daemon process')
        os.environ["SERVER_SOFTWARE"] = pulsar.SERVER_SOFTWARE
        pidfile = self.cfg.pidfile
        if pidfile is not None:
            p = Pidfile(pidfile)
            p.create(self.pid)
            self.pidfile = p
        PoolMixin.on_start(self)

    def periodic_task(self):
        # Arbiter periodic task
        if self.restarted:
            self.stop(exit_code=self.exit_code)
        if self.can_continue() and self.running():
            # managed actors job
            self.manage_actors()
            for m in list(itervalues(self.monitors)):
                if m.started():
                    if not m.running():
                        self.logger.info('Removing monitor %s', m)
                        self.monitors.pop(m.name)
                else:
                    m.start()
            try:
                self.cfg.arbiter_task(self)
            except:
                pass
        self.ioloop.add_callback(self.periodic_task, False)

    @async()
    def on_stop(self):
        '''Stop the pools the message queue and remaining actors.'''
        if not self.ioloop.running():
            for pool in list(itervalues(self.monitors)):
                pool.state = ACTOR_STATES.INACTIVE
            self.state = ACTOR_STATES.RUN
            self.restarted = True
            self.ioloop.start()
        else:
            self.state = ACTOR_STATES.STOPPING
            yield self.close_monitors()
            yield self.close_actors()
            yield self._close_message_queue()
    
    def on_exit(self):
        p = self.pidfile
        if p is not None:
            p.unlink()
        if self.managed_actors or self.linked_actors:
            self.state = ACTOR_STATES.TERMINATE
        self.logger.info("Bye.")
        if self.exit_code:
            sys.exit(self.exit_code)

    ############################################################################
    # INTERNALS
    ############################################################################
    def start(self):
        if self.state == ACTOR_STATES.INITIAL:
            if self.cfg.daemon: #pragma    nocover
                system.daemonize()
            return Actor.start(self)
            
    def _run(self):
        try:
            self.cfg.when_ready(self)
        except:
            pass
        try:
            self.requestloop.start()
        except HaltServer as e:
            self._halt(reason=str(e))
        except (KeyboardInterrupt, SystemExit) as e:
            self._halt(e.__class__.__name__)
        except:
            self._halt(code=1)

    def _halt(self, reason=None, code=None):
        if not self.closed():
            if not reason:
                self.logger.critical("Unhandled exception in main loop.",
                                     exc_info=True)
            else:
                self.logger.info(reason)
            self.stop(True, exit_code=code)

    def _close_message_queue(self):
        return

